import re


URL_PATTERN_STRING = "^(?P<scheme>[^:\/?#]+):\/{2,}(?P<path>[^?#]*)(\?([^#]*))?(#(.*))?$"
URL_PATTERN = re.compile(URL_PATTERN_STRING)
# PATTERN_METHOD_EXTRACTOR = re.compile(r"^(?P<module>(?:\w+\/)*\w+)(?:\.(?P<pattr>\w+)|(?::(?P<class>\w+)(?:\.(?P<iattr>\w+)|:(?P<cattr>\w+))?)?)?$")

PATTERN_METHOD_EXTRACTOR = re.compile(r"^(?P<module>(?:\w+\/)*\w+)(?P<attr>(?:\.\w+(?:\(\))?)*)?$")


def get_from_attr_segment(attr, seg):
    params_idx = seg.find('(')
    if params_idx > 0:
        # there is a () to indicate this attr should be called/instantiated
        attr = getattr(attr, seg[0:params_idx])()
    else:
        attr = getattr(attr, seg)
    return attr


import importlib
def get_from_attr_chain(module, attr_chain):

    attr = module
    for seg in attr_chain.strip('.').split('.'):
        attr = get_from_attr_segment(attr, seg)

    return attr


def get(url):
    '''

    urls like the following are understood.

    :param url:
    :return: The type of returned object depends on the url. here are the possibilities

        python://module                             # module
        python://pkg/module.attr                    # module attribute
        python://pkg/module:class                   # class
        python://pkg/module:class?arg1=x,argn=y     # class instance
        python://pkg/module:class.iattr             # class instance attribute
        python://pkg/module:class:cattr             # class static attribute


        someday support python versions and install package w version: 'python.2.7://requests[1.2,]|requests.get'
    '''
    m = URL_PATTERN.match(url)
    if not m:
        raise ValueError('url {} did not match pattern {}'.format(url,URL_PATTERN_STRING))

    d = m.groupdict()
    if not d['path']:
        raise ValueError('path in {} could not be extracted from {}'.format(URL_PATTERN_STRING, url))

    # The basic url has been parsed. now try to parse the path and args sections
    # to create a python thing
    m = PATTERN_METHOD_EXTRACTOR.match(d['path'])
    if not m:
        raise ValueError("path part of url ('{}') could not be parsed as python attribute chain".format(d['path']))

    # there must at least be a module
    d = m.groupdict()
    import importlib
    module = importlib.import_module(d['module'].replace('/','.'))

    if 'attr' in d and d['attr']:
        return get_from_attr_chain(module, d['attr'])

    #     # package attribute
    #     return getattr(module, d['pattr'])
    # if 'pattr' in d and d['pattr']:
    #     # package attribute
    #     return getattr(module, d['pattr'])
    #
    # elif 'class' in d and d['class']:
    #
    #     cls = getattr(module, d['class'])
    #     if 'iattr' in d and d['iattr']:
    #         # class instance method
    #         return getattr(cls(), d['iattr'])
    #
    #     if 'cattr' in d and d['cattr']:
    #         # class static method
    #         return cls.__getattribute__(d['cattr'])

    # the only remaining alternative is the module itself.
    else:
        return module


from . import url_scheme_python as pyurl
import jsonschema
import datetime
import six
import copy


class ModelRegistration(object):

    def __init__(self, model, data, fobj=None):

        self.factory = lambda: model.factory(copy.deepcopy(data))
        self.cached_object = None
        self.cached_since = None
        self.exception = None
        self.fobj = fobj

    def __repr__(self):
        return '<ModelRegistration {} {} {}>'.format(self.cached_object, self.cached_since, self.exception)

    def instance(self, reraise=False):

        if not self.cached_object:

            try:
                self.cached_object = self.factory()
                if self.fobj:
                    setattr(self.cached_object, '__flange', self.fobj)
                self.cached_since = datetime.datetime.now()
                self.exception = None
                return self.cached_object

            except Exception as e:
                self.exception = e
                if reraise:
                    raise e

        return self.cached_object


class Model(object):

    def __init__(self, name, validator, factory, inject=None):
        self.name = name
        self.factory = factory
        self.validator = validator
        self.inject = inject
        self.fobj = None

    def __repr__(self):
        return '<Model {}>'.format(self.name)


    def registration(self, data):
        return ModelRegistration(self, data, self.fobj)


    @staticmethod
    def get_schema_validator(schema):
        return lambda v: jsonschema.validate(v, schema) == None

    @staticmethod
    def from_plugin_config(config):

        if isinstance(config['schema'], six.string_types):
            # parse schema as a url of a method that returns the schema
            schema = pyurl.get(config['schema'])()
        else:
            # otherwise assume this is the schema itself as a dict
            schema = config['schema']

        return Model(
            config.get('name'),
            lambda v: jsonschema.validate(v, schema) == None,
            pyurl.get(config['factory']),
            config.get('inject'))



PLUGIN_SCHEMA = {
    'type': 'object',
    'properties': {
        'name': {'type':'string'},
        'type': {'constant': 'FLANGE.TYPE.PLUGIN'},
        'schema': {
            'oneOf': [
                {'type':'string', 'pattern': pyurl.URL_PATTERN_STRING},
                {'type': 'object'}]},
        'factory': {'type':'string', 'pattern': pyurl.URL_PATTERN_STRING},
        'inject': {"enum": ['flange']}  # if 'flange' then cause instances to have the '__flange' var set to the flange object
    },
    'required': ['name','type','schema','factory']}




logger_schema = {
    "type" : "object",
    "properties" : {
        'name': {'type':'string'},
        'level': { "enum": ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']},
        'format':{'type':'string'},
        'handler':{'type':'string'},
        'handler_args':{'type':'string'},
    },
    "required": ["name", "level"]
}


def logger_factory(config):
    import logging
    from logging import handlers
    log = logging.getLogger(config['name'])
    log.setLevel(getattr(logging, config['level'].upper()))
    if 'handler' in config:
        hargs = config['handler_args'] if 'handler_args' in config else {}
        try:
            # Create a handler assuming the class is defined in the logging.handlers class
            hdlr =  getattr(handlers, config['handler'])(**hargs)
        except:
            # Retry from the logging module directly
            hdlr =  getattr(logging, config['handler'])(**hargs)
    else:
        hdlr = logging.StreamHandler()
    if 'format' in config:
        hdlr.setFormatter(logging.Formatter(config['format']))
    else:
        hdlr.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s:%(name)s:%(message)s"))

    log.handlers = []
    log.addHandler(hdlr)
    return log




# A model to recognize other flange models in config
PLUGIN_MODEL = Model(
    'flange_plugin',
    Model.get_schema_validator(PLUGIN_SCHEMA),
    Model.from_plugin_config)


DEFAULT_MODELS = {

    # simple log model
    'logger': Model(
        'logger',
        Model.get_schema_validator(logger_schema),
        logger_factory)
}




import os
import anyconfig
from . import iterutils


PARSABLES = {
    'pickle':['p','pickle'],
    'xml':['xml'],
    'yaml':['yml','yaml'],
    'json':['json'],
    'ini':['ini'],
    'properties':['props','properties'],
    'shellvars':['env']}


class Source(object):

    def __init__(self, uri, root_path=None, contents={}, parser=None, error=None):
        self.uri = uri
        self.root_path = root_path
        self.error = error
        self.parser = parser
        self.contents = contents


    def __repr__(self):
        return self.__str__()
    def __str__(self):
        return "<Source uri={} root_path={} parser={} error={}>".format(self.uri, self.root_path, self.parser, self.error)



    def load(self):
        pass

        # try:
        #     self.parse()
        #
        # except Exception as e:
        #     # if the file had a known extension but didn't parse, raise an exception. The danger is that
        #     # it be parsed incorrectly as properties file which seems to match everything
        #     ext = os.path.splitext(self.uri)[1][1:]
        #     if [lext for lext in PARSABLES.values() if ext in lext]:
        #         self.error = e
        #         # print type(e) # 'exception parsing {}\t{}'.format(ext, e)
        #
        #     for p in PARSABLES.keys():
        #         try:
        #             self.parse(p)
        #             break
        #         except Exception as e:
        #             # print type(e) #'exception parsing as ', p, ' ', e
        #             pass



    # def parse(self, parser=None):
    #
    #     self.contents = anyconfig.load(self.uri, ac_parser=parser, ac_ordered=True)
    #     self.parser = parser if parser else os.path.splitext(self.uri)[1].strip('.')



    @staticmethod
    def from_file(full_file_path, root_path):
        s = SourceFile(full_file_path, root_path)
        s.load()
        return s


#
# class SourceDict(Source):
#
#     def __init__(self, uri, root_path=None, contents={}, parser=None, error=None):
#         super(SourceDict, self).__init__(uri, root_path, contents, parser, error)



class SourceFile(Source):

    # def __init__(self, uri, root_path=None, contents={}, parser=None, error=None):
    #     super(SourceFile, self).__init__(uri, root_path, contents, parser, error)


    def _parse(self, parser=None):

        contents = anyconfig.load(self.uri, ac_parser=parser, ac_ordered=True)
        parser = parser if parser else os.path.splitext(self.uri)[1].strip('.')
        return (contents, parser)


    def load(self):

        try:
            self.contents, self.parser = self._parse()

        except Exception as e:
            # if the file had a known extension but didn't parse, raise an exception. The danger is that
            # it be parsed incorrectly as properties file which seems to match everything
            ext = os.path.splitext(self.uri)[1][1:]
            if [lext for lext in PARSABLES.values() if ext in lext]:
                self.error = e
                # print type(e) # 'exception parsing {}\t{}'.format(ext, e)

            for p in PARSABLES.keys():
                try:
                    self.contents, self.parser = self._parse(p)
                    break
                except Exception as e:
                    # print type(e) #'exception parsing as ', p, ' ', e
                    pass





import requests
# import jmespath



def get_model_instance(data):
    auth = (data['user'], data['password']) if 'user' in data else None
    r = requests.get(data['url'], verify=False, auth=auth).json()
    # if 'jmespath' in config:
    #     r = jmespath.search(config['jmespath'], r)
    return r


FACTORY = get_model_instance

SCHEMA = {
    "type" : "object",
    "properties" : {
        'url':{'type':'string'},
        'user':{'type':'string'},
        'password':{'type':'string'},
        'extract':{'type':'string'},
    },
    "required": ["url"]
}





import os, fnmatch, string, six, pprint
from . import iterutils, model as flmd
from .source import Source, SourceFile
import anyconfig
from collections import OrderedDict


DEFAULT_EXCLUDE_PATTERNS = ['*.tar','*.jar','*.zip','*.gz','*.swp','node_modules','target','.idea','*.hide','*save']
DEFAULT_FILE_PATTERNS = ['*.yml','*cfg','*settings','*config','*properties','*props']
VALID_KEY_CHARS = [c for c in string.printable if c not in ['_'] ]
DEFAULT_KEY_FILTER = lambda p, k, v: k == None or isinstance(k, int) or (isinstance(k, six.string_types) and len(k)<50 and all(c in string.printable for c in VALID_KEY_CHARS))
DEFAULT_UNFLATTEN_SEPARATOR = '__'



def from_home_dir(root_path=None, include_os_env=False):
    return Cfg(base_dir='~', root_path=root_path, include_os_env=include_os_env)

def from_file(filename, root_path=None, include_os_env=False):
    path, basename = os.path.split(os.path.realpath(os.path.expanduser(filename)))
    return Cfg(root_path=root_path, include_os_env=include_os_env, base_dir=path, file_patterns=[basename])

def from_dict(d, root_path=None, include_os_env=False):
    return Cfg(data=d, root_path=root_path, file_patterns=None, include_os_env=include_os_env)

def from_os_env(root_path=None):
    return Cfg(root_path=root_path, include_os_env=True, file_patterns=[])




#
#   Provide a Global object for convenience. Shadow the main class methods at the module level.
#

__GLOBAL_CFG = None
def __GET_GLOBAL_FLANGE():
    global __GLOBAL_CFG
    if not __GLOBAL_CFG:
        __GLOBAL_CFG = Cfg()
    return __GLOBAL_CFG

def obj(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().obj(*args, **kwargs)
def objs(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().objs(*args, **kwargs)
def path(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().path(*args, **kwargs)
def paths(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().paths(*args, **kwargs)
def src(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().src(*args, **kwargs)
def srcs(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().srcs(*args, **kwargs)
def uri(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().uri(*args, **kwargs)
def uris(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().uris(*args, **kwargs)
def value(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().value(*args, **kwargs)
def values(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().values(*args, **kwargs)
def fobj(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().fobj(*args, **kwargs)
def fobjs(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().fobjs(*args, **kwargs)
def search(*args, **kwargs):
    return __GET_GLOBAL_FLANGE().search(*args, **kwargs)


def info(path=None):
    return __GET_GLOBAL_FLANGE().info(path=path)


def register_model(name, model, research=True):
    return __GET_GLOBAL_FLANGE().register_model(name, model, research)



def register_default_model(name, model):

    flmd.DEFAULT_MODELS[name] = model

    # If the global instance is already constructed, then explicitly register and research now
    if __GLOBAL_CFG:
        return register_model(name, model)





class PathCacheObject(object):

    def __init__(self, val=None, mregs=None, srcs=None, path=None, wrap=False):

        if wrap:
            # basic wrapper. untested. keep this around for experimentation. This could be used to
            # keep the metadata this class holds in-place in the original data via substitution of the
            # wrapped object. Then an index of paths wouldn't be necessary.
            #
            # copied from:
            #   https://stackoverflow.com/questions/1443129/completely-wrap-an-object-in-python/1445289#1445289
            self.__class__ = type(val.__class__.__name__,
                                  (self.__class__, val.__class__),
                                  {})
            self.__dict__ = val.__dict__


        self.val = val
        self.mregs = mregs if mregs else {}
        self.srcs = srcs if srcs else set()
        self.path = path




    @staticmethod
    def from_path_and_value(pv):
        return PathCacheObject(path=pv[0], val=pv[1])

    def __repr__(self):
        return "<PathCacheObject {}, #srcs={}, #instances={}>".format(self.path, len(self.srcs), len(self.mregs))


    def add_src(self, src):
        self.srcs.add(src)

    def add_model(self, model, v):
        # print 'add model called on ',  model
        if model.name not in self.mregs:
            self.mregs[model.name] = model.registration(v)


    def val_equals(self, val):
        # print 'assert_val_equals: ', self.val, val
        return self.val == val


    def instance(self, model=None, reraise=True):

        if model:
            if model in self.mregs:
                # try:
                return self.mregs[model].instance(reraise=reraise)
                # except Exception as e:
                #     print e


        elif len(self.mregs) == 1:
            return list(self.mregs.values())[0].instance(reraise=reraise)

        if len(self.mregs) > 1:
            raise ValueError('multiple registrations. Must specify model name from :{}'.format(self.mregs))





class Cfg(object):
    
    def __init__(self, 
                data=None,
                include_os_env=True,
                root_path=None,
                base_dir='.',
                file_patterns=DEFAULT_FILE_PATTERNS,
                file_exclude_patterns=DEFAULT_EXCLUDE_PATTERNS,
                file_search_depth=0,
                unflatten_separator=DEFAULT_UNFLATTEN_SEPARATOR,
                key_filter=DEFAULT_KEY_FILTER,
                src_post_proc=None,
                gather=True,
                load=True,
                merge=True,
                research=True,):
        """

        :param data: initial data. This is merged as is without regard to root_path
        :param include_os_env:
        :param research_models:
        :param root_path: the namespace/key under which to add all loaded config/data. If model instances are defined at top level this will be needed
        :param base_dir: directory or list of directories to search for config/data. ** order matters! later entries override earlier.
        :param file_patterns:
        :param file_exclude_patterns:
        :param file_search_depth:
        :param unflatten_separator:
        """

        # save params
        self.unflatten_separator = unflatten_separator
        self.file_patterns = [file_patterns] if isinstance(file_patterns, six.string_types) else file_patterns
        self.file_exclude_patterns = file_exclude_patterns
        self.file_search_depth = file_search_depth
        self.include_os_env = include_os_env
        self.root_path = root_path
        self.init_data = data
        # self.gather = gather
        # self.merge = merge
        # self.research = research

        self.key_filter = key_filter
        self.src_post_proc = src_post_proc

        self.base_dir = base_dir
        if isinstance(self.base_dir, six.string_types):
            self.base_dir = [self.base_dir]

        # init data
        self.data = {}
        self.sources = []
        self.path_index = {}
        self.models = flmd.DEFAULT_MODELS.copy()

        # function to give to Source objects so register themselves with the path_index
        # self.source_indexer = lambda src, p, k, v: self.__visit_index_path(self.path_index, src, p, k, v)


    # conditionally do initial gather, merge and research
        self.refresh(gather, load, merge, research)


    #
    #
    #   Flange Cfg() init methods
    #
    #

    def clear_data(self):

        if self.data:
            del self.data
            self.data = {}
        if self.path_index:
            del self.path_index
            self.path_index = {}
        if self.models:
            del self.models
            self.models = flmd.DEFAULT_MODELS.copy()


    def refresh(self, gather=False, load=True, merge=True, research=True):

        clear = False
        if gather:
            self.gather_sources()
            clear = True
        if load:
            self.load_sources()
            clear = True

        if clear:
            self.clear_data()

        if merge:
            self.merge_sources()
        if research:
            self.research_models()



    def gather_sources(self):

        if self.sources:
            del self.sources

        self.visited_uris = set()


        if self.file_patterns:
            for dir in self.base_dir:
                self.sources.extend(self.__get_file_sources(dir))

        if self.include_os_env:
            # Dont use root path
            self.sources.append(Source('os_env', '', os.environ.copy()))

        if self.init_data:
            # Dont use root path
            self.sources.append(Source('init_data', '', self.init_data))



    def load_sources(self):
        for s in self.sources:
            s.load()


    def merge_sources(self):

        # process sources with called provided function. This gives the caller a chance to
        # shape things up or set the src path prior to the filter, index and merge
        dlist = []
        for s in self.sources:
            if self.src_post_proc:
                self.src_post_proc(s)

            d = {s.root_path: s.contents} if s.root_path else s.contents
            # print d.keys()
            e = iterutils.unflatten(d, self.unflatten_separator)
            # print 'after unflatten', e
            dlist.append(iterutils.remap(e, reraise_visit=True, enter=lambda p, k, v: self.__filter_and_index(s, p, k, v)))
            # e =  {'test': {'exclude': [['192.168.0.0/16'], ['172.16.0.0/12', '10.0.0.0/8'], '10.1.3.0/24']}}
            # dlist.append(iterutils.remap(e, enter=lambda p, k, v: self.__filter_and_index(s, p, k, v)))


        # then merge, putting the content under the root path for each source
        self.data = {}
        failed = []
        for d in dlist:
            try:
                # print 'merging ', d
                anyconfig.merge(self.data, d)
            except:
                failed.append(d)




    def research_models(self):

        plugins = iterutils.research(
            self.data,
            query=lambda p, k, v: flmd.PLUGIN_MODEL.validator(v))

        for p in plugins:
            m = self.register_model(
                p[1]['name'],
                flmd.PLUGIN_MODEL.factory(p[1]),
                False)
            # Give self to the plugin model instances so, in turn, the models can provide to their instances
            if m.inject == 'flange':
                m.fobj = self

        iterutils.research(
            self.data,
            query=lambda p, k, v: self.__visit_index_model_instance(self.models.values(), p, k, v))





    #
    #
    #   Primary access methods
    #
    #



    def search(self, path, values=None, unique=False, raise_absent=False, vfunc=lambda x: x):
        """
       Return single model object instance matching given criteria
       :param path: tuple or dpath expression representing the hierarchy/chain of parent keys
       :param values: single value or list of values to match. If exact is False then .contains method is used as filter
       :param raise_absent: if True then raise exception if no match is found
       :return: list matching ojects directly from data/config in the form of ((k1, k2, .., kn), value)
       """
        path_and_value_list = iterutils.search(
                self.data,
                path=path,
                required_values=values)

        # print 'search found ', [x[0] for x in path_and_value_list]

        return self.__return_value(path_and_value_list, unique, raise_absent, vfunc)



    def __return_value(self, l, unique=True, raise_absent=False, vfunc=lambda x: x):

        if l:
            if vfunc:
                l = [y for y in [vfunc(x) for x in l] if y]

            if len(l) > 1 and unique:
                raise ValueError('multiple matches found')

        if not l:
            if raise_absent:
                raise ValueError('no match found')

        else:
            return l[0] if unique else l



    def path(self, path=None, values=None, raise_absent=False):
        return self.search(path=path, unique=True, values=values, raise_absent=raise_absent, vfunc=lambda x: x[0])
    def paths(self, path=None, values=None, raise_absent=False):
        return self.search(path=path, unique=False, values=values, raise_absent=raise_absent, vfunc=lambda x: x[0])


    def src(self, path=None, values=None, raise_absent=False):
        sources = self.search(path=path, values=values, unique=True, raise_absent=raise_absent, vfunc=lambda x: self.path_index[x[0]].srcs)
        return next(iter(sources))
    def srcs(self, path=None, values=None, raise_absent=False):
        sources = self.search(path=path, values=values, unique=False, raise_absent=raise_absent, vfunc=lambda x: self.path_index[x[0]].srcs)
        return list(set([s for l in sources for s in l])) if sources else sources



    def uri(self, path=None, values=None, raise_absent=False):
        sources = self.search(path=path, unique=True, values=values, raise_absent=raise_absent, vfunc=lambda x: self.path_index[x[0]].srcs)
        return next(iter(sources)).uri
    def uris(self, path=None, values=None, raise_absent=False):
        sources = self.search(path=path, unique=False, values=values, raise_absent=raise_absent, vfunc=lambda x: self.path_index[x[0]].srcs)
        return [src.uri for l in sources for src in l] if sources else sources



    def obj(self, path=None, model=None, values=None, raise_absent=False):
        """
       Return single model object instance matching given criteria
       :param path: tuple or dpath expression representing the hierarchy/chain of parent keys
       :param values: single value or list of values to match. If exact is False then .contains method is used as filter
       :param raise_absent: if True then raise exception if no match is found
       :return: matching object from cache if already created or new if not
       """
        return self.search(path=path, unique=True, raise_absent=raise_absent, values=values,
                            vfunc=lambda x: self.path_index[x[0]].instance(model=model) if x[0] in self.path_index else None)
    def objs(self, path=None, model=None, values=None, raise_absent=False):
        """
       Return list of model object instances matching given criteria
       :param path: tuple or dpath expression representing the hierarchy/chain of parent keys
       :param values: single value or list of values to match. If exact is False then .contains method is used as filter
       :param raise_absent: if True then raise exception if no match is found
       :return: list of matching objects
       """
        return self.search(path=path, unique=False, raise_absent=raise_absent, values=values,
                       vfunc=lambda x: self.path_index[x[0]].instance(model=model, reraise=False) if x[0] in self.path_index else None)


    def value(self, path=None, values=None, raise_absent=False):
        """
        Return single data/config value matching given criteria
        :param path: tuple or dpath expression representing the hierarchy/chain of parent keys
        :param values: single value or list of values to match. If exact is False then .contains method is used as filter
        :param raise_absent: if True then raise exception if no match is found
        :return: matching value
        """
        return self.search(path=path, unique=True, values=values, raise_absent=raise_absent, vfunc=lambda x: x[1])

    def values(self, path=None, values=None, raise_absent=False):
        """
        Return all data/config values matching given criteria
        :param path: tuple or dpath expression representing the hierarchy/chain of parent keys
        :param values: single value or list of values to match. If exact is False then .contains method is used as filter
        :param raise_absent: if True then raise exception if no match is found
        :return: list of matching values
        """
        return self.search(path=path, unique=False, values=values, raise_absent=raise_absent, vfunc=lambda x: x[1])


    def fobj(self, path=None, values=None, unique=True, raise_absent=False):
        """
        Return model instance/registration object matching given criteria
        :param path: tuple or dpath expression representing the hierarchy/chain of parent keys
        :param values: single value or list of values to match. If exact is False then .contains method is used as filter
        :param raise_absent: if True then raise exception if no match is found
        :return: single model instance/registration object
        """
        return self.path_index[self.search(path=path, unique=unique, values=values, raise_absent=raise_absent)[0]]

    def fobjs(self, path=None, values=None, raise_absent=False):
        """
        Return all model instance/registration objects matching given criteria
        :param path: tuple or dpath expression representing the hierarchy/chain of parent keys
        :param values: single value or list of values to match. If exact is False then .contains method is used as filter
        :param raise_absent: if True then raise exception if no match is found
        :return: list of model instance/registration objects
        """
        return self.search(path=path, unique=False, values=values, raise_absent=raise_absent,
                           vfunc=lambda x: self.path_index[x[0]] if x[0] in self.path_index else None)



    #
    #
    #   Init helper methods and remap callbacks
    #
    #


    # def register_model_schema(self, name, schema, factory, research=True):
    #     self.register_model(
    #         name,
    #         flmd.Model(name, flmd.Model.get_schema_validator(schema), factory),
    #         research)

    def register_model(self, name, model, research=True):
        self.models[name] = model
        if research:
            iterutils.research(
                self.data,
                query=lambda p, k, v: self.__visit_index_model_instance([model], p, k, v))
        return model


    def __filter_and_index(self, src, p, k, v):

        # now the root_path/ns has already been accounted for in the data.
        # no prefixing loginc
        # np = (src.root_path,) + p if src.root_path else p
        np = p


        # print '__filter_and_index', p, k
        # filter. func may be provided or be class default
        if self.key_filter and not self.key_filter(np, k, v):
            # print 'filtered out ', np, k
            return v, False

        # index. internal to Cfg class
        full_path = np + (k,)
        if full_path in self.path_index:
            # print 'preexisting index at ', full_path
            if not self.path_index[full_path].val_equals(v):
                # print 'updating index at ', full_path
                # raise ValueError('unexpected value change at path_index[{}]'.format(full_path))
                self.path_index[full_path].add_src(src)
        else:
            # print 'adding index at ', full_path
            self.path_index[full_path] = PathCacheObject(val=v, path=full_path, srcs=set([src]))


        return iterutils.default_enter(p, k, v)


    def __visit_index_model_instance(self, models, p, k, v):
        """
            Called during model research on merged data
        """
            # print 'model visit {} on {}'.format(model, v)
        cp = p + (k,)
        for model in models:
            try:
                if model.validator(v):
                    if cp in self.path_index:
                        # if self.path_index[cp].val != v:
                        #     raise ValueError('unexpected value change at path_index[{}]'.format(cp))
                        self.path_index[cp].add_model(model, v)
                    else:
                        # The object should already be in the index but don't complain for now.
                        self.path_index[cp] = PathCacheObject(val=v, path=cp, regs=[model])
            except:
                pass



    def __get_file_sources(self, topdir):

        sources = []
        topdir = os.path.realpath(os.path.expanduser(topdir))
        start_depth = topdir.count(os.sep)

        # print '__get_file_sources on ', dir
        for root, dirnames, filenames in os.walk(os.path.realpath(os.path.expanduser(topdir)), topdown=True):

            # print 'walk: {}. {}. {}'.format(root, dirnames, filenames)

            depth = root.count(os.sep) - self.base_dir.count(os.sep) - start_depth
            if depth >= self.file_search_depth:
                # print 'depth {} exceeded {} on {}'.format(depth, self.file_search_depth, root)
                del dirnames[:]


            # remove excluded directories and files
            for e in self.file_exclude_patterns:
                for d in [dirname for dirname in fnmatch.filter(dirnames, e)]:
                    dirnames.remove(d)
                for f in [filename for filename in fnmatch.filter(filenames, e)]:
                    # print 'removing ', f
                    filenames.remove(f)

            to_include = set()
            for p in self.file_patterns:
                for filename in fnmatch.filter(filenames, p):
                    to_include.add(os.path.join(root, filename))

            for filename in to_include:
                # print filename
                # Don't add a uri twice
                # print '{} is in {}: {}'.format(filename,  self.visited_uris, filename in self.visited_uris)
                if filename not in self.visited_uris:

                    src = SourceFile(filename, self.root_path)
                    if src:
                        sources.append(src)
                        self.visited_uris.add(src.uri)

        return sources




    #
    #
    #
    #
    #


    def info(self, path=None):


        def iprint(s, level=0):
            s = str(s)
            print(s.rjust(len(s)+3*level))

        def psrcs(srcs, level=0):
            # print('\n')
            iprint('sources:', level)
            for src in srcs:
                iprint("{0:15.10} {1:60.65} {2}".format(str(src.parser), str(src.uri), 'error: ' + str(src.error) if src.error else ''), level+1)


        def pmodels(flobjs, omit_empty=False, level=0):
            # print('\n')
            iprint('models:', level)
            for model in self.models:
                mobjs = [x for x in flobjs if x.mregs.get(model)]
                if not omit_empty or mobjs:
                    # print('\n')
                    iprint('{}'.format(model), level+1)
                    for x in mobjs:
                        iprint("{0:50} {1}".format('/'.join(x.path), x.mregs.get(model)), level+2)

        def pvalues(values, level=0):
            # print('\n')
            iprint('values:', level)
            vs = [dict(x) if isinstance(x, OrderedDict) else x for x in values]
            if vs:
                for v in vs:
                    iprint(v, level+1)


        s = ''
        if not path:
            print('\nconfig:\nbase dir: \t{}\nsearch depth: \t{}\nfile include patterns: \t{}\nfile exclude patterns: \t{}'.format(
                self.base_dir , self.file_search_depth, self.file_patterns, self.file_exclude_patterns))
            print('\n')
            pmodels(self.path_index.values())
            print('\n')
            psrcs(self.sources)
        else:

            for k in search(path):
                print('\n{}:'.format('/'.join(k[0])))
                pmodels(fobjs(k[0]), omit_empty=True, level=1)

                psrcs(srcs(path), level=1)

                pvalues(values(path), level=1)












from . import cfg, model
from sqlalchemy import create_engine

dbengine_schema = {
    "type" : "object",
    "properties" : {
        'driver':{'type':'string'},
        'name':{'type':'string'},
        'user':{'type':'string'},
        'pass':{'type':'string'},
        'port':{'type':'string'},
    },
    "required": ["driver", "name", "user", "pass"]
}


def dbengine_create_func(config):

    url = "{:s}://{:s}:{:s}@{:s}:{:s}/{:s}?".format(
        config.pop('driver'),
        config.pop('user'),
        config.pop('pass'),
        config.pop('host'),
        config.pop('port'),
        config.pop('name'))

    return create_engine(url, **config)


def register():
    cfg.register_default_model(
        'dbengine',
        model.Model('dbengine',
                    model.Model.get_schema_validator(dbengine_schema),
                    dbengine_create_func))

register()



import logging
from .context import *
from nose.tools import *

from flange import dbengine, iterutils, url_scheme_python as pyurl
# from flange import Flange, url_scheme_python as pyurl

# pyurl = url_scheme_python

#.
#   url scheme
#



def get_module_function_test_value():
    return 'module function'

module_name = __name__.replace('.','/')


def test_url_module_function():
    attr = pyurl.get('python://{}.get_module_function_test_value'.format(module_name))
    assert attr() == 'module function'

    attr = pyurl.get('python://{}.get_module_function_test_value()'.format(module_name))
    assert attr == 'module function'


def test_url_class():
    attr = pyurl.get('python://{}.TestPlugin'.format(module_name))
    assert attr().get_schema__instance()

    attr = pyurl.get('python://{}.TestPlugin()'.format(module_name))
    assert attr.get_schema__instance()


def test_url_class_functions():
    attr = pyurl.get('python://{}.TestPlugin().get_schema__instance'.format(module_name))
    assert attr()['type'] == 'object'
    attr = pyurl.get('python://{}.TestPlugin().get_schema__instance()'.format(module_name))
    assert attr['type'] == 'object'

    attr = pyurl.get('python://{}.TestPlugin.get_schema__static'.format(module_name))
    assert attr()['type'] == 'object'
    attr = pyurl.get('python://{}.TestPlugin.get_schema__static()'.format(module_name))
    assert attr['type'] == 'object'



#
#   InstanceRegistry
#



#
#   flange api
#
data = {
    'a':{'a2':{'a3a':'a3aval', 'a3b':'a3aval'}},
    'testlog': {
        'name': 'testlog',
        'level': 'DEBUG',
        'format': "%(asctime)s:%(levelname)s:%(name)s:%(message)s"},
    'testlog2': {
        'name': 'another name',
        'level': 'WARNING',
        'format': "%(asctime)s:%(levelname)s:%(name)s:%(message)s"}
}

f = flange.cfg.Cfg(data=data)

#
#   registry = f.register(schema)
#
#   src_meta = f.source(url_or_file_path, validation_model)
#
#   f.validate(
#




def test_search_key_exact():
    assert f.search('a/a2/a3b')
    assert f.search('a/a2/a3*')
    assert not f.search('a/a2/a3')
    assert not f.search('a/a2/a3b_')


@raises(ValueError)
def test_search_key_exact_raise():
    f.search('3b', raise_absent=True)


def test_search_key_fuzzy():
    assert f.search('**/a3b*')
    assert len(f.search('**/a3*', unique=False)) == 2
    assert not f.search('a/a2/a3bx*')
    assert not f.search('a/a2/*XX*')


@raises(ValueError)
def test_multiples_raise_when_unique_specified():
    # even with 'p' given, unique means unique
    f.search('**/a3*', unique=True)


@raises(ValueError)
def test_search_key_fuzzy_raise():

    assert not f.search('**/XX', raise_absent=True)





def test_model_research():
    assert type(f.obj('testlog')) == logging.Logger


def test_obj_by_name():
    assert(f.obj('testlog'))
    assert(f.obj('testlog2'))
    assert(f.obj('testlog') != f.obj('testlog2'))


def test_obj_by_model():
    # An instance can be fetched by model name. If there are multiple instances a values can be used
    with assert_raises(ValueError):
        f.obj(model='logger', raise_absent=True)
    assert(f.obj(model='logger', values='DEBUG'))


def test_obj_values():
    # values can be used with or without the config key to retrieve an instance
    assert(f.obj('testlog', values='DEBUG'))
    assert(f.obj('testlog', values='WARNING') == None)

    assert(f.obj('testlog2', values='WARNING'))
    assert(f.obj('testlog2', values='DEBUG') == None)



def test_iter_search():


    # r = iterutils.__query(p, v, k, accepted_keys=None, required_values=None, path=None, exact=True)

    # key
    assert iterutils.__query(('root',), 'testkey', None,
                             'testkey', None, None, True)
    assert iterutils.__query(('root',), 'testkey', None,
                             ['testkey', 'xxx'], None, None, True)
    assert iterutils.__query(('root',), 'testkey', None,
                             'test', None, None, False)
    assert iterutils.__query(('root',), 'testkey', None,
                             ['test', 'xxx'], None, None, False)
    assert not iterutils.__query(('root',), 'testkey', None,
                                 'test', None, None, True)
    assert not iterutils.__query(('root',), 'testkey', None,
                                 'testkeyextra', None, None, True)
    assert not iterutils.__query(('root',), 'testkey', None,
                                 'testkeyextra', None, None, False)

    # path
    assert iterutils.__query(('root',), 'testkey', None,
                             'testkey', None, ('root',), True)
    assert not iterutils.__query(('root',), 'testkey', None,
                                 'testkey', None, ('root', 'sub',), True)

    # val
    assert iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                             None, 'val1', None, False)
    assert iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                             None, 'val2', None, False)
    assert iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                             None, 'val1', None, True)
    assert iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                             None, 'va', None, False)
    assert iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                             None, '2', None, False)
    assert iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                             None, ['val1', 'val2'], None, True)
    assert not iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                                 None, 'v3', None, False)
    assert not iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                                 None, ['v3'], None, False)
    assert not iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                                 None, 'val', None, True)
    assert not iterutils.__query(('root',),  'testkey', {'k1':'val1', 'k2':'val2'},
                             None, ['val1', 'val2', 'val3'], None, True)



def test_get():

    po = f.fobj('testlog')
    assert po



class TestPlugin:

    class Inner:
        def m(self):
            pass

    '''
        A class to test providing schema and factory with instance methods, as well
        a test the created object which is an instance of this class. so both the model
        and the instance are provided by this class
    '''

    def __init__(self, value=None):
        self.value = value

    def get_value(self):
        return self.value

    @staticmethod
    def get_schema__static():
        return {
            'type': 'object',
            'properties':{
                'only_TestPlugin_would_match_this': {'type': 'string'}
            },
            'required': ['only_TestPlugin_would_match_this']
        }

    def get_schema__instance(self):
        return TestPlugin.get_schema__static()


    def get_instance(self, params):
        # return an instance of this object with the value given in params
        return TestPlugin(params['only_TestPlugin_would_match_this'])



def test_plugin_model():


    data = {

        'test_instance_key': {
            'only_TestPlugin_would_match_this': 'some value'
        },

        'test_plugin_config_key': {
            'name': 'test_plugin',
            'type': 'FLANGE.TYPE.PLUGIN',
            'schema': {
                'type': 'object',
                'properties':{
                    'only_TestPlugin_would_match_this': {'type': 'string'}
                },
                'required': ['only_TestPlugin_would_match_this']
            },
            'factory': 'python://{}.TestPlugin().get_instance'.format(module_name)
        }
    }

    f = flange.cfg.Cfg(data=data, file_patterns=None)
    # print f.obj('test_instance_key')
    assert f.obj('test_instance_key').get_value() == 'some value'



def test_sqlalchemy_plugin():

    try:
        dbengine.register()
    except ImportError as e:
        # It's ok if sqlalchemy can't be found. Its not required
        assert('sqlalchemy' in str(e))


#
#   File handling
#


# def test_get_file_source():
#     fn = 'test/resources/dsh_example.yml'
#     s = flange.Flange.(fn, ns='testns', ns_from_dirname=True)
#     assert fn in s['src']
#     assert s['dict']
#     assert s['ns'] == 'testns'

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))

PATH_EXAMPLE = os.path.join(os.path.dirname(__file__), 'resources', 'dsh_example.yml')

#
def test_instantiate_from_explicit_file():
    f = flange.cfg.from_file(PATH_EXAMPLE, root_path='contexts__default')
    assert f.search('contexts/default/vars')


def test_delayed_merge():
    f = flange.cfg.Cfg(base_dir=os.path.dirname(__file__), file_search_depth=1, merge=False)
    assert not f.search('command_name2')
    f.refresh(False, True, True)
    assert f.search('command_name2')



import anyconfig
import iterutils


class Data(object):

    def __init__(
            self,
            unflatten_separator="__",
            indexed_obj_factory=None):
        """

        :param unflatten_separator:
        :param indexed_obj_factory: func(p, k, v, existing_obj) - the factory function to
        create an object to be stored at the each path in the index. If there is an existing
        object it is passed to this function so that the caller/func can decide how to
        hanlde multiple values at a given path
        """

        self.data = {}
        self.path_index = {}
        self.unflatten_separator = unflatten_separator
        self.indexed_obj_factory = indexed_obj_factory


    def add(self, d, root_path):

        # try:
        d = {root_path: d} if root_path else d
        e = iterutils.unflatten(d, self.unflatten_separator)
        anyconfig.merge(self.data, iterutils.remap(e, lambda p, k, v: self.__filter_and_index(p, k, v)))
        # except:
        #     self.failed.append(d)



    def search(self, path_expression, mode=UXP, values=None, ifunc=lambda x: x):
        """
        find matches for the given path expression in the data

        :param path_expression: path tuple or string
        :return:
        """
        # keys = path_expression if isinstance(path_expression, six.string_types) else path_expression[-1]

        path_and_value_list = iterutils.search(
            self.data,
            path_expression=path_expression,
            required_values=values,
            exact=(mode[1] == "x"))

        return self.__return_value(path_and_value_list, mode, ifunc)



    def __return_value(self, l, mode, ifunc=lambda x: x):

        if l:
            if len(l) > 1 and mode[0] == 'u':
                raise ValueError('multiple matches found')

        elif mode[2] == 'r':
            raise ValueError('no match found')


        if mode[0] == 'u':
            return ifunc(l[0]) if l else None
        else:
            return l if not ifunc else [ifunc(x) for x in l]


    def __visit_index_path(self, src, p, k, v):
        """
            Called during processing of source data
        """
        cp = p + (k,)
        self.path_index[cp] = self.indexed_obj_factory(p, k, v, self.path_index.get(cp))
        if cp in self.path_index:
            # if self.path_index[cp].assert_val_equals(v):
            #     raise ValueError('unexpected value change at path_index[{}]'.format(cp))
            self.path_index[cp].add_src(src)
        else:
            self.path_index[cp] = Flobject(val=v, path=cp, srcs=set([src]))


    def __filter_and_index(self, p, k, v):

        # now the root_path/ns has already been accounted for in the data.
        # no prefixing loginc
        # np = (src.ns,) + p if src.ns else p
        np = p

        # filter. func may be provided or be class default
        if self.key_filter and not self.key_filter(np, k, v):
            return False

        # index. internal to Cfg class
        full_path = np + (k,)
        if full_path in self.path_index:
            if not self.path_index[full_path].val_equals(v):
                print 'updating index at ', full_path
                # raise ValueError('unexpected value change at path_index[{}]'.format(full_path))
                self.path_index[full_path].add_src(src)
        else:
            print 'adding index at ', full_path
            self.path_index[full_path] = indexed_obj_factory(p, k, v)

        return k, v



# https://packaging.python.org/tutorials/distributing-packages/

from setuptools import setup, find_packages

readme = open('README.md', 'r')
README_TEXT = readme.read()
readme.close()

setup(
    name='flange',
    version='0.0.8',
    author='flashashen',
    author_email='flashashen@gmail.com',
    description='Autoload configuration from multiple sources, object registry, dpath access.',
    license = "MIT",
    url="https://github.com/flashashen/flange",
    classifiers= [
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Environment :: Console',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Development Status :: 3 - Alpha',

    ],
    platforms='osx,linux,mswindows',
    keywords = "configuration yaml object registry spring",
    long_description=README_TEXT,
    packages=find_packages(exclude=['test']),
    py_modules=['flange'],
    install_requires=[
        'anyconfig',
        'PyYAML',
         'six',
        'jsonschema',
        'dpath'
    ],
)